/*
 * Copyright 2014-2016 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.testbase;

import ru.ispras.fortress.solver.SolverId;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.testbase.generator.DataGenerator;

import java.util.Collection;
import java.util.Collections;

public final class TestBase {
  private static TestBase instance = null;
  private static SolverId solverId = SolverId.Z3_TEXT;

  private final TestBaseRegistry registry;

  public static TestBase get() {
    if (null == instance) {
      instance = new TestBase();
    }
    return instance;
  }

  public static SolverId getSolverId() {
    return solverId;
  }

  public static void setSolverId(SolverId value) {
    InvariantChecks.checkNotNull(solverId);
    TestBase.solverId = value;
  }

  private TestBase() {
    this.registry = new TestBaseRegistry();
  }

  public TestBaseRegistry getRegistry() {
    return registry;
  }

  public TestBaseQueryResult executeQuery(final TestBaseQuery query) {
    try {
      return executeQueryInternal(query);
    } catch (final Throwable e) {
      return newErrorResult(e);
    }
  }

  private TestBaseQueryResult executeQueryInternal(final TestBaseQuery query) {
    InvariantChecks.checkNotNull(query);

    final String name = (String) query.getContext().get(TestBaseContext.TESTCASE);
    final Collection<DataGenerator> generators = registry.getNamedGenerators(name);

    if (generators != null) {
      for (final DataGenerator generator : generators) {
        if (generator.isSuitable(query)) {
          return TestBaseQueryResult.success(generator.generate(query));
        }
      }
    }

    return TestBaseQueryResult.NOTFOUND;
  }

  private static TestBaseQueryResult newErrorResult(final Throwable e) {
    final StringBuilder sb = new StringBuilder();

    sb.append("Data generation failed. Reason: ");
    sb.append(String.format("Exception: %s: %s", e.getClass().getName(), e.getMessage()));
    sb.append(System.lineSeparator());

    final java.io.StringWriter writer = new java.io.StringWriter();
    e.printStackTrace(new java.io.PrintWriter(writer));
    sb.append(writer.toString());

    return TestBaseQueryResult.reportErrors(
        Collections.singletonList(sb.toString()));
  }
}
